@@ -12,9 +12,8 @@ import dash
 
 # Enable enrich as drop-in replacement for dash
 # noinspection PyUnresolvedReferences
-from dash import (
+from dash import (  # lgtm [py/unused-import]
     no_update,
-    Input,
     Output,
     State,
     ClientsideFunction,
@@ -32,7 +31,7 @@ from dash import (
     clientside_callback,
 )
 from dash._utils import patch_collections_abc
-from dash.dependencies import _Wildcard
+from dash.dependencies import _Wildcard, DashDependency  # lgtm [py/unused-import]
 from dash.development.base_component import Component
 from flask import session
 from flask_caching.backends import FileSystemCache, RedisCache
@@ -41,9 +40,20 @@ from collections import defaultdict
 from typing import Dict, Callable, List, Union, Any, Tuple
 from datetime import datetime
 
+from dash_extensions import CycleBreaker
+
 _wildcard_mappings = {ALL: "<ALL>", MATCH: "<MATCH>", ALLSMALLER: "<ALLSMALLER>"}
 _wildcard_values = list(_wildcard_mappings.values())
 
+# region Enriched dependencies
+
+class Input(dash.Input):
+    def __init__(self, component_id, component_property, break_cycle=False):
+        super().__init__(component_id, component_property)
+        self.break_cycle = break_cycle
+
+# endregion
+
 # region Dash blueprint
 
 CbInput = Union[State, Input]
@@ -213,6 +223,9 @@ class DashProxy(dash.Dash):
         self.blueprint.register_callbacks(super())
         # Proceed as normally.
         super()._setup_server()
+        # Remap callback bindings to enable callback registration via the 'before_first_request' hook.
+        self.callback = super().callback
+        self.clientside_callback = super().clientside_callback
         # Set session secret. Used by some subclasses.
         if not self.server.secret_key:
             self.server.secret_key = secrets.token_urlsafe(16)
@@ -222,7 +235,8 @@ class DashProxy(dash.Dash):
         Hijack another app. Typically, used with Dataiku 10 where the Dash object is instantiated outside user code.
         """
         # Change properties.
-        app.config.update(self.config)
+        readonly_props = app.config.__dict__.get("_read_only", {})
+        app.config.update({k: v for k, v in self.config.items() if k not in readonly_props})
         app.title = self.title
         app.index_string = self.index_string
         # Inject layout.
@@ -345,9 +359,15 @@ class BlockingCallbackTransform(DashTransform):
             start_client_id = f"{callback_id}_start_client"
             end_server_id = f"{callback_id}_end_server"
             end_client_id = f"{callback_id}_end_client"
-            self.components.extend(
-                [dcc.Store(id=start_client_id), dcc.Store(id=end_server_id), dcc.Store(id=end_client_id)]
-            )
+            start_blocked_id = f"{callback_id}_start_blocked"
+            end_blocked_id = f"{callback_id}_end_blocked"
+            self.components.extend([
+                dcc.Store(id=start_client_id),
+                dcc.Store(id=end_server_id),
+                dcc.Store(id=end_client_id),
+                dcc.Store(id=start_blocked_id),
+                CycleBreaker(id=end_blocked_id)
+            ])
             # Bind start signal callback.
             start_callback = f"""function()
             {{
@@ -355,48 +375,61 @@ class BlockingCallbackTransform(DashTransform):
                 const end = arguments[arguments.length-1];
                 const now = new Date().getTime();
                 if(!end & !start){{
-                    return now;
+                    return [now, null];
                 }}
-                if((now - start)/1000 > {timeout}){{              
-                    console.log("HITTING TIMEOUT");  
-                    return now;
+                if((now - start)/1000 > {timeout}){{
+                    // timed out
+                    return [now, null];
                 }}
                 if(!end){{
-                    return window.dash_clientside.no_update;
+                    // blocked
+                    return [window.dash_clientside.no_update, now];
                 }}
                 if(end > start){{
-                    return now;
+                    return [now, null];
                 }}
-                return window.dash_clientside.no_update;
+                return [window.dash_clientside.no_update, now];
             }}"""
             self.blueprint.clientside_callback(
                 start_callback,
-                Output(start_client_id, "data"),
-                callback.inputs,
+                [Output(start_client_id, "data"), Output(start_blocked_id, "data")],
+                callback.inputs + [Input(end_blocked_id, "dst")],
                 [State(start_client_id, "data"), State(end_client_id, "data")],
             )
             # Bind end signal callback.
+            end_callback = """function(endServerId, startBlockedId)
+            {
+                const now = new Date().getTime();
+                if(startBlockedId){
+                    return [now, now];
+                }
+                return [now, window.dash_clientside.no_update];
+            }"""
             self.blueprint.clientside_callback(
-                "function(){return new Date().getTime();}", Output(end_client_id, "data"), Input(end_server_id, "data")
+                end_callback,
+                [Output(end_client_id, "data"), Output(end_blocked_id, "src")],
+                Input(end_server_id, "data"),
+                State(start_blocked_id, "data")
             )
             # Modify the original callback to send finished signal.
+            single_output = len(callback.outputs) <= 1
             callback.outputs.append(Output(end_server_id, "data"))
             # Modify the original callback to not trigger on inputs, but the new special trigger.
             new_state = [State(item.component_id, item.component_property) for item in callback.inputs]
             callback.inputs = [Input(start_client_id, "data")] + new_state
             # Modify the callback function accordingly.
             f = callback.f
-            callback.f = skip_input_signal_add_output_signal()(f)
+            callback.f = skip_input_signal_add_output_signal(single_output)(f)
 
         return callbacks
 
 
-def skip_input_signal_add_output_signal():
+def skip_input_signal_add_output_signal(single_output: bool):
     def wrapper(f):
         @functools.wraps(f)
         def decorated_function(*args):
             value = f(*args[1:])
-            return _as_list(value) + [datetime.utcnow().timestamp()]
+            return _as_output_list(value, single_output) + [datetime.utcnow().timestamp()]
 
         return decorated_function
 
@@ -508,11 +541,12 @@ class LogTransform(DashTransform):
             if not callback.kwargs.get("log", None):
                 continue
             # Add the log component as output.
+            single_output = len(callback.outputs) <= 1
             callback.outputs.append(self.log_config.log_output)
             # Modify the callback function accordingly.
             f = callback.f
             logger = DashLogger(self.log_config.log_writer_map)  # TODO: What about scope?
-            callback.f = bind_logger(logger)(f)
+            callback.f = bind_logger(logger, single_output)(f)
 
         return callbacks
 
@@ -520,19 +554,59 @@ class LogTransform(DashTransform):
         return [MultiplexerTransform()]
 
 
-def bind_logger(logger):
+def bind_logger(logger, single_output):
     def wrapper(f):
         @functools.wraps(f)
         def decorated_function(*args):
             logger.clear()
             value = f(*args, logger)
-            return _as_list(value) + [logger.get_output()]
+            return _as_output_list(value, single_output) + [logger.get_output()]
 
         return decorated_function
 
     return wrapper
 
 
+# endregion
+
+# region Cycle breaker transform
+
+class CycleBreakerTransform(DashTransform):
+
+    def __init__(self):
+        super().__init__()
+        self.components = []
+
+    def transform_layout(self, layout):
+        children = _as_list(layout.children) + self.components
+        layout.children = children
+
+    def apply(self, callbacks, clientside_callbacks):
+        cycle_inputs = {}
+        # Update inputs.
+        for c in callbacks + clientside_callbacks:
+            for i in c.inputs:
+                if isinstance(i, Input) and i.break_cycle:
+                    cid = self._cycle_break_id(i)
+                    cycle_inputs[cid] = (i.component_id, i.component_property)
+                    i.component_id = cid
+                    i.component_property = "dst"
+        # Construct components.
+        self.components = [CycleBreaker(id=cid) for cid in cycle_inputs]
+        # Construct callbacks.
+        f = "function(x){return x;}"
+        cycle_callbacks = []
+        for cid in cycle_inputs:
+            cb = CallbackBlueprint(Output(cid, "src"), Input(*cycle_inputs[cid]))
+            cb.f = f
+            cycle_callbacks.append(cb)
+        return callbacks, clientside_callbacks + cycle_callbacks
+
+    @staticmethod
+    def _cycle_break_id(d: DashDependency):
+        return f"{str(d).replace('.', '_')}_breaker"
+
+
 # endregion
 
 # region Global blueprint object (to emulate Dash 2.0 import syntax)
@@ -642,6 +716,7 @@ def dynamic_prefix(app: Union[DashBlueprint, DashProxy], component: Component):
     prefix_transform: PrefixIdTransform = prefix_transforms[0]
     prefix_component(prefix_transform.prefix, component, prefix_transform.escape)
 
+
 # endregion
 
 # region Trigger transform (the only default transform)
@@ -1087,6 +1162,8 @@ class Dash(DashProxy):
             LogTransform(),
             MultiplexerTransform(),
             NoOutputTransform(),
+            CycleBreakerTransform(),
+            BlockingCallbackTransform(),
             ServersideOutputTransform(**output_defaults),
         ]
         super().__init__(*args, transforms=transforms, **kwargs)
@@ -1106,6 +1183,12 @@ def _as_list(item):
     return [item]
 
 
+def _as_output_list(item, single_output: bool):
+    if single_output:
+        return [item]
+    return _as_list(item)
+
+
 def _create_callback_id(item):
     cid = item.component_id
     if isinstance(cid, dict):
